/*
	Matthias Hollnder & Tamy Boubekeur
	Copyright (c) 2010 Telecom ParisTech
*/

// --------------------------------------------------
// shader definition/extensions
// --------------------------------------------------
#version 120
#extension GL_EXT_gpu_shader4 : enable
#extension GL_ARB_texture_rectangle : enable				// The BasisFunctionTable-Texture is as Rectangular texture

// --------------------------------------------------
// varying variables
// --------------------------------------------------
varying vec3 			V;									// Vertex in View-Space
varying vec3 			N;									// Normal of the Vertex
varying vec3 			vLightDir;							// Vec To Light

// --------------------------------------------------
// uniform variables:
// --------------------------------------------------
uniform int				VerticesSamplerSize;
uniform sampler2D		VerticesSampler;					// those are the vertices of the model subdivided once via GPU
uniform sampler2D		NormalsSampler;						// those are the normals of the model subdivided once via GPU

uniform int				PatchSamplerSize;
uniform usampler2D		PatchSampler;						// The patches of the model
uniform int				iOneRingSize;						// how many uvec4 are for the One-Ring-Neighbourhood
uniform int				iPatchSize;							// how many uvec4 are for the whole Patch

uniform sampler2DRect	BFTSampler;							// The BasisFunctionTables Rectangular format
uniform int				BFTSamplerWidth;					// Width and Height
uniform int				iTableSize;							// how many vec4 one BasisFunctionTable comprises
uniform int				iMaxValenceGenerated;				// for which MaxValence the BFT was generated

uniform bool			bDrawNormals;

// --------------------------------------------------
// const (see GeometryEngine.h)
// --------------------------------------------------
const unsigned int		uiVertexTag_Interior 	= 0u;
const unsigned int		uiVertexTag_Border 		= 1u;
const unsigned int		uiVertexTag_Corner 		= 2u;
const unsigned int		uiVertexTag_Crease 		= 3u;

const unsigned int		uiMaxValence = 20u;

// --------------------------------------------------
// function to convert 1D-index to 2D-index
// Normed between 0..1
// --------------------------------------------------
vec2 Convert1DIndexTo2DIndex_Normed(in unsigned int uiVertexIndex, in int iWrapSize)
{
		float fWrapRec = 1.0 / float(iWrapSize);
		unsigned int uiWrap = unsigned int(iWrapSize);
		unsigned int uiY = uiVertexIndex / uiWrap;
		unsigned int uiX = uiVertexIndex - (uiY* uiWrap);
		
		float fX = float(uiX) * fWrapRec;
		float fY = float(uiY) * fWrapRec;
		
		return vec2(fX, fY);
}

// --------------------------------------------------
// function to convert 1D-index to 2D-index
// Unnormed between 0..w, 0..h
// --------------------------------------------------
ivec2 Convert1DIndexTo2DIndex_Unnormed(in unsigned int uiVertexIndex, in int iWrapSize)
{
		unsigned int uiWrap = unsigned int(iWrapSize);
		unsigned int uiY = uiVertexIndex / uiWrap;
		unsigned int uiX = uiVertexIndex - (uiY* uiWrap);

		return ivec2( int( uiX ), int( uiY ) );
}

// --------------------------------------------------
// helper function to convert the patchnumber to an
// index that can be used for usampler Patches
// --------------------------------------------------
unsigned int GetPatchIndex(unsigned int uiPatchNumber)
{
		return uiPatchNumber * (unsigned int(iPatchSize));
}

// --------------------------------------------------
// functions for calculating how many BFTs exist
// for a certain type
// --------------------------------------------------
unsigned int GetNumBasisFunction_Corner()
{
		return 1u;
}
unsigned int GetNumBasisFunction_Special()
{
		return 1u;
}
unsigned int GetNumBasisFunction_Interior()
{
		return unsigned int(iMaxValenceGenerated-2);
}
unsigned int GetNumBasisFunction_Border()
{
		return unsigned int (((iMaxValenceGenerated-3)*(iMaxValenceGenerated-2)) / 2);
}
unsigned int GetNumBasisFunction_Crease()
{
		return unsigned int(iMaxValenceGenerated-2);
}
unsigned int GetNumBasisFunction_Crease2()
{
		return unsigned int(iMaxValenceGenerated-2);
}

// --------------------------------------------------
// Functions for retrieving BasisFunctionOffset
// --------------------------------------------------
unsigned int GetOffsetCornerTables()
{
		return 0u;
}

unsigned int GetOffsetSpecialTables()
{
		return 1u;
}

unsigned int GetOffsetInteriorTables(unsigned int uiValence)
{
		return ((GetNumBasisFunction_Corner() + GetNumBasisFunction_Special() + (uiValence-3u)));
}

unsigned int GetOffsetCreaseTables(unsigned int uiValence0, unsigned int uiValence1)
{ 
		if (uiValence1 == 4u) {
			return ((GetNumBasisFunction_Corner() + GetNumBasisFunction_Special() + GetNumBasisFunction_Interior() + uiValence0-3u));
		}
		else {
			return ((GetNumBasisFunction_Corner() + GetNumBasisFunction_Special() + GetNumBasisFunction_Interior() + GetNumBasisFunction_Crease() + uiValence1-3u));
		}
}
unsigned int GetOffsetBorderTables(unsigned int uiValence, unsigned int uiPosition)
{ 
		unsigned int uiShift = (((uiValence-4u)*(uiValence-3u))/2u);
		return (GetNumBasisFunction_Corner() + GetNumBasisFunction_Special() + GetNumBasisFunction_Interior() + GetNumBasisFunction_Crease() + GetNumBasisFunction_Crease2() + uiShift + uiPosition-1u);
}

// --------------------------------------------------
// calculate which basis-function to use
// --------------------------------------------------
unsigned int GetBFTIndex(in uvec4 uvValenceAndPosition, in uvec4 uvTagsAndOneRingSize)
{
		if (uvTagsAndOneRingSize.x == uiVertexTag_Corner)
		{
				return GetOffsetCornerTables();
		}
		else if (uvTagsAndOneRingSize.x == uiVertexTag_Interior)
		{
				return GetOffsetInteriorTables(uvValenceAndPosition.x);
		}
		else if (uvTagsAndOneRingSize.x == uiVertexTag_Crease)
		{
				return GetOffsetCreaseTables(uvValenceAndPosition.x, uvValenceAndPosition.y);
		}
		else if (uvTagsAndOneRingSize.y == uiVertexTag_Border)
		{
				return GetOffsetSpecialTables();
		}
		else
		{
				return GetOffsetBorderTables(uvValenceAndPosition.x, uvValenceAndPosition.w);
		}
}

// --------------------------------------------------
// Applying the BasisFunctionTable to a Vertex
// @param uiPatchIndex is the index of the current patch
// @param uiBFTIndex is the index of the BFT to use
// @param uiOneRingSizePatch is the size of the one-ring for this patch
// --------------------------------------------------
void ApplyBasisFunction( in unsigned int uiPatchIndex, in unsigned int uiBFTIndex, in unsigned int uiOneRingSizePatch, out vec4 vVecOut, out vec3 vNormalOut )
{
		vec4 vReturnVertex = vec4(0.0, 0.0, 0.0, 0.0);
		vec3 vReturnNormal = vec3(0.0, 0.0, 0.0);
		
		vec4 vWeights = vec4(0.0, 0.0, 0.0, 0.0);
		uvec4 vCtrlPointsIndices = uvec4(0u,0u,0u,0u);
		vec4 vVertex = vec4(0.0, 0.0, 0.0, 0.0);
		vec3 vNormal = vec3(0.0, 0.0, 0.0);
		
		unsigned int uiCtrlIndex = 0u;
		unsigned int uiCurrentWeightIndex = 0u;
		vec2 vTexCoord = vec2(0.0, 0.0);
		ivec2 vTexCoord_Unnormed = ivec2(0, 0);
		unsigned int uiOneRingSizeCounter = 0u;																	// This is a counter of how many patch-vertices have been used
		unsigned int uiVertexIndex = 0u;
		
		for (unsigned int i=0u; i<uiMaxValence; ++i)
		{
				// fetch 4 weights
				uiCurrentWeightIndex = uiBFTIndex + i;															// index of the next 4 weights
				vTexCoord_Unnormed = Convert1DIndexTo2DIndex_Unnormed(uiCurrentWeightIndex, BFTSamplerWidth);	// texture coordinate of the 4 weights
				vWeights = texture2DRect(BFTSampler, vTexCoord_Unnormed);										// look-up 4 weights
				
				// fetch Indices of 4 ctrl-points
				uiCtrlIndex = uiPatchIndex + i;
				vTexCoord = Convert1DIndexTo2DIndex_Normed(uiCtrlIndex, PatchSamplerSize);
				vCtrlPointsIndices = texture2D(PatchSampler, vTexCoord);
				
				
				bool bHasFinished = false;
				
				////////////////////////////////////////
				// sum up values
				////////////////////////////////////////
				for (int j=0; j<4; ++j)
				{
						vTexCoord = Convert1DIndexTo2DIndex_Normed(vCtrlPointsIndices.x, VerticesSamplerSize);	// fetching texcoord for vertex
						vVertex = texture2D(VerticesSampler, vTexCoord);										// retrieving vertex by texcoord
						vReturnVertex += vVertex * vWeights.x;													// adding weighted vertex
						
						vNormal = texture2D(NormalsSampler, vTexCoord).xyz;										// retrieving normal by texcoord (same texcoord as vertex)
						vReturnNormal += vNormal * vWeights.x;													// adding weighted normal
						++uiOneRingSizeCounter;
						
						if (uiOneRingSizeCounter >= uiOneRingSizePatch) {
							bHasFinished = true;
							break;
						}
						// shuffle the values
						vWeights = vWeights.yzwx;
						vCtrlPointsIndices = vCtrlPointsIndices.yzwx;						
				}
				if (bHasFinished) {
					break;
				}
		}
		vVecOut = vReturnVertex;
		vNormalOut = normalize(vReturnNormal);
}

// --------------------------------------------------
// Vertex-Shader
// --------------------------------------------------
void main() {
		// Get the patch by InstanceID
		
		unsigned int uiPatchIndex = GetPatchIndex( unsigned int ( gl_InstanceID ) );			// get the patch-index
		unsigned int uiIndexValence = uiPatchIndex + unsigned int( iOneRingSize );				// The index for the Valence
		
		vec2 vTexCoord = Convert1DIndexTo2DIndex_Normed( uiIndexValence, PatchSamplerSize );	// Texcoord for the valence
		uvec4 uvValenceAndPosition = texture2D( PatchSampler, vTexCoord );						// this will fetch the valence and position tags of the patch
		vTexCoord = Convert1DIndexTo2DIndex_Normed( uiIndexValence+1u, PatchSamplerSize );		// Texcoord for the Tags and OneRingSize
		uvec4 uvTagsAndOneRingSize = texture2D( PatchSampler, vTexCoord );						// this will fetch the VertexTags and OneRingsize of the patch

		unsigned int uiVertexIndex = unsigned int ( gl_Vertex.x );
		
		unsigned int uiBFTIndex = GetBFTIndex( uvValenceAndPosition, uvTagsAndOneRingSize );
		uiBFTIndex *= unsigned int(iTableSize);
		uiBFTIndex += uiVertexIndex * ( ( unsigned int ( iMaxValenceGenerated ) + 6u ) /4u );
		
		vec4 vPos 		= vec4( 0.0, 0.0, 0.0, 0.0 );
		vec3 vNormal 	= vec3( 0.0, 0.0, 0.0 );
		ApplyBasisFunction( uiPatchIndex, uiBFTIndex, uvTagsAndOneRingSize.w, vPos, vNormal );
		
		N = ( gl_NormalMatrix * vNormal );

		V = ( gl_ModelViewMatrix * vPos ).xyz;
		vLightDir = gl_LightSource[0].position.xyz - V;
		gl_Position = ( gl_ModelViewProjectionMatrix * vPos );
}

